#include "def.h"
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>

char disp_line[LINE_NUM_MAX][LINE_MAX];
global_option option = {13, 0, 0, 0, 0, 4, MODE_ALPHA, BLENDED, "", DEFAULT_FONT_PATH, {0, 0, 0, 0}, {255, 255, 255, 0}};

SDL_Surface *gScreenSurface;
SDL_Surface *TextSurface;
TTF_Font *font;

bool Init(void);
void End(void);
bool LoadFile_kanji(void);
bool LoadFile_alpha(void);
int LoadBookmark(const char *f);
bool SaveBookmark(const char *f, int num);
bool MainLoop(void);
void DrawBar(int start, int height);
void GetInput(void);
void FillScreen(SDL_Color color);
void Usage(void);

bool Init(void) {
  FILE *fp;

  if (SDL_Init(SDL_INIT_VIDEO) < 0) return false;
  //  if (SDL_Init(SDL_INIT_EVERYTHING) < 0) return false;

  //  if (SDL_EnableKeyRepeat(300, 50)) return false;  

  gScreenSurface = SDL_SetVideoMode(
				    SCREEN_WIDTH,
				    SCREEN_HEIGHT,
				    SCREEN_BPP,
				    SDL_SWSURFACE// | SDL_FULLSCREEN
				    );

  SDL_ShowCursor(SDL_DISABLE);

  if (TTF_Init()) return false;
  fp = fopen(option.fontfile, "r");
  if (!fp) return false;
  fclose(fp);
  font = TTF_OpenFont(option.fontfile, option.font_size);

  option.line_height = TTF_FontHeight(font) + option.line_space;
  option.disp_line_max = HEIGHT_MAX/option.line_height - 1;

  return true;
}

void End(void) {
  //  SDL_FreeSurface(TextSurface);
  if (font) TTF_CloseFont(font);

  SDL_Quit();
}

bool LoadFile_kanji(void) {
  FILE *fp;
  char buf[BUF_MAX];
  char *src_p, *dst_p;
  int line_idx=0, bytes=1, i;
  int width=0, narrow_char=0, wide_char=0;

  printf("Opening file: %s\n", option.filename);
  fp = fopen(option.filename, "r");
  if (!fp) return false;

  TTF_SizeUTF8(font, "a", &narrow_char, NULL);
  TTF_SizeUTF8(font, "あ", &wide_char, NULL);  
  printf("font width: %d, %d\n", narrow_char, wide_char);

  dst_p = &disp_line[0][0];
  while (fgets(buf, sizeof(buf), fp)) {
    src_p = &buf[0];
    while (*src_p != '\0' && *src_p != '\n') {
      if ((unsigned char)*src_p <= 0x7f) {
	bytes = 1;
	width+=narrow_char;
      } else if ((unsigned char)*src_p <= 0xbf) {
	printf("error: splitted letter\n");
	return false;
      } else if ((unsigned char)*src_p <= 0xdf) {
	bytes = 2;
	width+=narrow_char;
      } else if ((unsigned char)*src_p <= 0xef) {
	bytes = 3;
	width+=wide_char;
      } else if ((unsigned char)*src_p <= 0xf7) {
	bytes = 4;
	width+=wide_char;
      } else {
	printf("error: unknown letter\n");
	return false;
      }

      if (width > WIDTH_MAX) {
	if (line_idx+1 >= LINE_NUM_MAX) {
	  printf("error: too long text\n");
	  return false;
	}
	*dst_p = '\0';
	line_idx++;
	dst_p = &disp_line[line_idx][0];
	switch (bytes) {
	case 1:
	case 2:
	  width=narrow_char;
	  break;
	case 3:
	case 4:
	  width=wide_char;
	}
      }

      for (i=0; i<bytes; i++) {
	if (*src_p == '\0') {
	  fgets(buf, sizeof(buf), fp);
	  src_p = &buf[0];
	}
	*dst_p = *src_p;
	src_p++; dst_p++;
      }
    }
    if (*src_p == '\n') {
      *dst_p = '\0';
      line_idx++;
      if (line_idx >= LINE_NUM_MAX) {
	printf("error: too long text\n");
	return false;
      }
      dst_p = &disp_line[line_idx][0];
      width=0;
    }
  }
  fclose(fp);
  option.line_max = line_idx-1;
  printf("line: %d\n", option.line_max+1);

  option.bar_size = 240 * (option.disp_line_max+1)/line_idx;
  if (option.bar_size < 4) option.bar_size = 4;

  return true;
}

bool LoadFile_alpha(void) {
  FILE *fp;
  char buf[BUF_MAX];
  char *src_p, *dst_p, *last_space;
  int width=0, line_idx=0, bytes=1, i;
  struct stat st;
  int progress_block, processed_size=0, processed_block=0;
  SDL_Rect r;

  stat(option.filename, &st);
  progress_block = (int)st.st_size/20;
  r.x = 65; r.y = 95;
  r.w = 190; r.h = 20;
  SDL_FillRect(gScreenSurface, &r, SDL_MapRGB(gScreenSurface->format, 100, 100, 255));
  r.x = 70; r.y = 100;
  r.w = 180; r.h = 10;
  SDL_FillRect(gScreenSurface, &r, SDL_MapRGB(gScreenSurface->format, 50, 50, 50));
  SDL_Flip(gScreenSurface);

  printf("Opening file: %s\n", option.filename);
  fp = fopen(option.filename, "r");
  if (!fp) return false;

  dst_p = &disp_line[0][0];
  last_space = dst_p;
  while (fgets(buf, sizeof(buf), fp)) {
    src_p = &buf[0];
    while (*src_p != '\0' && *src_p != '\n') {
      if ((unsigned char)*src_p <= 0x7f) {
	bytes = 1;
      } else if ((unsigned char)*src_p <= 0xbf) {
	printf("error: splitted letter\n");
	return false;
      } else if ((unsigned char)*src_p <= 0xdf) {
	bytes = 2;
      } else if ((unsigned char)*src_p <= 0xef) {
	bytes = 3;
      } else if ((unsigned char)*src_p <= 0xf7) {
	bytes = 4;
      } else {
	printf("error: unknown letter\n");
	return false;
      }
      if (*src_p == ' ') {
	last_space = dst_p;
      }

      for (i=0; i<bytes; i++) {
	if (*src_p == '\0') {
	  fgets(buf, sizeof(buf), fp);
	  src_p = &buf[0];
	}
	*dst_p = *src_p;
	src_p++; dst_p++;
	*dst_p = '\0';
      }

      processed_size += bytes;
      if (processed_size > progress_block) {
	r.x = 70+9*processed_block; r.y = 100;
	r.w = 9; r.h = 10;
	SDL_FillRect(gScreenSurface, &r, SDL_MapRGB(gScreenSurface->format, 255, 255, 255));
	SDL_Flip(gScreenSurface);
	processed_block++;
	processed_size = 0;
      }

      TTF_SizeUTF8(font, disp_line[line_idx], &width, NULL);
      if (width > WIDTH_MAX) {
	if (line_idx+1 >= LINE_NUM_MAX) return false;
	if (last_space == &disp_line[line_idx][0]) {
	  for (i=0; i<bytes; i++) {
	    disp_line[line_idx+1][i] = *(dst_p-bytes+i);
	  }
	  dst_p -= bytes;
	  *dst_p = '\0';
	  line_idx++;
	  dst_p = &disp_line[line_idx][bytes];
	  last_space = &disp_line[line_idx][0];
	} else {
	  line_idx++;
	  dst_p = &disp_line[line_idx][0];
	  *last_space = '\0';
	  for (last_space = last_space+1; *last_space != '\0'; last_space++) {
	    *dst_p = *last_space;
	    dst_p++;
	  }
	  last_space = &disp_line[line_idx][0];
	}
      }
    }
    if (*src_p == '\n') {
      *dst_p = '\0';
      line_idx++;
      if (line_idx >= LINE_NUM_MAX) {
	printf("error: too long text\n");
	return false;
      }
      dst_p = &disp_line[line_idx][0];
      last_space = dst_p;
    }
  }
  fclose(fp);
  option.line_max = line_idx-1;
  printf("line: %d\n", option.line_max+1);

  option.bar_size = 240 * (option.disp_line_max+1)/line_idx;
  if (option.bar_size < 4) option.bar_size = 4;

  return true;
}

int LoadBookmark(const char *f) {
  FILE *fp;
  char buf[BUF_MAX];
  int num=0;

  fp = fopen(BOOKMARK_PATH, "r");
  if (!fp) return 0;

  while (fgets(buf, sizeof(buf), fp)) {
    if (strstr(buf, f) == NULL) continue;
    if (fgets(buf, sizeof(buf), fp)) {
      num = atoi(buf);
      break;
    }
  }

  fclose(fp);
  return num;
}

bool SaveBookmark(const char *f, int num) {
  FILE *ifp, *ofp;
  char buf[BUF_MAX];
  bool found=false;

  ifp = fopen(BOOKMARK_PATH, "r");
  if (!ifp) {
    ofp = fopen(BOOKMARK_PATH, "w");
    fprintf(ofp, f);
    fprintf(ofp, "\n%d\n", num);
    fclose(ofp);
  } else {
    ofp = fopen(BOOKMARK_TMP_PATH, "w");
    while (fgets(buf, sizeof(buf), ifp)) {
      fprintf(ofp, buf);
      if (strstr(buf, f) != NULL) {
	fgets(buf, sizeof(buf), ifp);
	fprintf(ofp, "%d\n", num);
	found = true;
      }
    }
    if (!found) {
      fprintf(ofp, f);
      fprintf(ofp, "\n%d\n", num);
    }
    fclose(ifp); fclose(ofp);
    remove(BOOKMARK_PATH);
    rename(BOOKMARK_TMP_PATH, BOOKMARK_PATH);
  }

  return true;
}

bool MainLoop(void) {
  SDL_Event ev;
  SDL_Rect dest;
  int i, current_line, bar_start=0;
  bool redraw_flag = true;

  current_line = LoadBookmark(option.filename);
  if (current_line > option.line_max) current_line = 0;

  while (1) {
    SDL_Delay(10);

    SDL_PollEvent(&ev);
    switch (ev.type){
    case SDL_QUIT:
      SaveBookmark(option.filename, current_line);
      return true;
      break;
    }

    GetInput();
    if (button_to_process[PRESS_BTN_SELECT]) {
      SaveBookmark(option.filename, current_line);
      return true;
    }
    if (button_to_process[PRESS_BTN_DOWN]) {
      if (current_line+1 <= option.line_max) {
	current_line++;
	redraw_flag = true;
      }
    }
    if (button_to_process[PRESS_BTN_RIGHT] || button_to_process[PRESS_BTN_A]) {
      if (current_line+option.disp_line_max <= option.line_max) {
	current_line += option.disp_line_max;
	redraw_flag = true;
      }
    }
    if (button_to_process[PRESS_BTN_UP]) {
      if (current_line >= 1) {
	current_line--;
	redraw_flag = true;
      }
    }
    if (button_to_process[PRESS_BTN_LEFT]) {
      current_line -= option.disp_line_max;
      if (current_line < 0) current_line = 0;
      redraw_flag = true;
    }
    if (button_to_process[PRESS_BTN_L]) {
      if (current_line != 0) {
	current_line = 0;
	redraw_flag = true;
      }
    }
    if (button_to_process[PRESS_BTN_R]) {
      if (current_line < option.line_max-option.disp_line_max+1) {
	current_line = option.line_max-option.disp_line_max+1;
	redraw_flag = true;
      }
    }

    if (redraw_flag) {
      FillScreen(option.bg_color);
      for (i=0; i<=option.disp_line_max; i++) {
	switch (option.rendering_func) {
	case SOLID:
	  TextSurface = TTF_RenderUTF8_Solid(font, disp_line[current_line+i], option.fg_color);
	  break;
	case SHADED:
	  TextSurface = TTF_RenderUTF8_Shaded(font, disp_line[current_line+i], option.fg_color, option.bg_color);
	  break;
	case BLENDED:
	  TextSurface = TTF_RenderUTF8_Blended(font, disp_line[current_line+i], option.fg_color);
	  break;
	default:
	  break;
	}
	dest.x = 2;
	dest.y = 2+i*(option.line_height);
	SDL_BlitSurface(TextSurface, NULL, gScreenSurface, &dest);
	SDL_FreeSurface(TextSurface);
      }

      bar_start = 240*current_line/option.line_max;
      if (bar_start > 240-option.bar_size) bar_start = 240-option.bar_size;
      DrawBar(bar_start, option.bar_size);

      SDL_Flip(gScreenSurface);

      redraw_flag = false;
    }

  }

  return false;
}

void DrawBar(int start, int height) {
  SDL_Rect dest;

  dest.x = 315; dest.y = 0;
  dest.w = 4;
  dest.h = 240;
  //  SDL_FillRect(gScreenSurface, &dest, SDL_MapRGB(gScreenSurface->format, 200, 200, 200));
  SDL_FillRect(gScreenSurface, &dest, 0xCE59);
  dest.x = 315; dest.y = start;
  dest.w = 4;
  dest.h = height;
  SDL_FillRect(gScreenSurface, &dest, 0x00000000);
}

void GetInput(void) {
  static int delay[PRESS_BTN_MAX] = {REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY, REPEAT_DELAY};
  static int current_btn[PRESS_BTN_MAX] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  static int prev_btn[PRESS_BTN_MAX];
  int i;
  Uint8 *keys;

  for (i=0; i<PRESS_BTN_MAX; i++) {
    prev_btn[i] = current_btn[i];
    current_btn[i] = 0;
  }
  keys = SDL_GetKeyState(NULL);
  current_btn[PRESS_BTN_UP]     = keys[SDLK_UP];
  current_btn[PRESS_BTN_DOWN]   = keys[SDLK_DOWN];
  current_btn[PRESS_BTN_LEFT]   = keys[SDLK_LEFT];
  current_btn[PRESS_BTN_RIGHT]  = keys[SDLK_RIGHT];
  current_btn[PRESS_BTN_A]      = keys[SDLK_LCTRL];
  current_btn[PRESS_BTN_B]      = keys[SDLK_LALT];
  current_btn[PRESS_BTN_X]      = keys[SDLK_SPACE];
  current_btn[PRESS_BTN_Y]      = keys[SDLK_LSHIFT];
  current_btn[PRESS_BTN_L]      = keys[SDLK_TAB];
  current_btn[PRESS_BTN_R]      = keys[SDLK_BACKSPACE];
  current_btn[PRESS_BTN_START]  = keys[SDLK_RETURN];
  current_btn[PRESS_BTN_SELECT] = keys[SDLK_ESCAPE];

  for (i=0; i<PRESS_BTN_MAX; i++) {
    if (current_btn[i]) {
      if (prev_btn[i]) {
	delay[i]--;
	if (delay[i] <= 0) {
	  button_to_process[i] = true;
	  delay[i] = REPEAT_INTERVAL;
	} else {
	  button_to_process[i] = false;
	}
      } else {
	button_to_process[i] = true;
	delay[i] = REPEAT_DELAY;
      }
    } else {
      button_to_process[i] = false;
      delay[i] = REPEAT_DELAY;
    }
  }

  return;
}

void FillScreen(SDL_Color color) {
  SDL_Rect dest;

  dest.x = 0; dest.y = 0;
  dest.w = SCREEN_WIDTH;
  dest.h = SCREEN_HEIGHT;
  SDL_FillRect(gScreenSurface, &dest, SDL_MapRGB(gScreenSurface->format, color.r, color.g, color.b));
}

void Usage(void) {
  fprintf(stderr,"Usage: easytext [option] textfile\n");
  fprintf(stderr," -a           : \"alphabet\" mode\n");
  fprintf(stderr," -k           : \"kanji\" mode\n");
  fprintf(stderr," -s #         : font size\n");
  fprintf(stderr," -l #         : line space\n");
  fprintf(stderr," -r #         : rendering function\n");
  fprintf(stderr," -f fontfile  : font file\n");
}

int main(int argc, char *argv[]) {
  int ch;
  extern char *optarg;
  extern int optind, opterr;
  SDL_Rect dest;
  
  while ((ch = getopt(argc, argv, "kas:l:r:f:")) != -1){
    switch (ch){
    case 'k':
      printf("kanji mode\n");
      option.text_mode = MODE_KANJI;
      break;
    case 'a':
      printf("alphabet mode\n");
      option.text_mode = MODE_ALPHA;
      break;
    case 's':
      printf("font size: %d\n",atoi(optarg));
      option.font_size = atoi(optarg);
      if (option.font_size > 30 || option.font_size < 0) option.font_size = 12;
      break;
    case 'l':
      printf("line space: %d\n",atoi(optarg));
      option.line_space = atoi(optarg);
      if (option.line_space > 30 || option.line_space < 0) option.line_space = 1;
      break;
    case 'r':
      printf("rendering function: %d\n",atoi(optarg));
      switch (atoi(optarg)) {
      case 0:
	option.rendering_func = SOLID;
	break;
      case 1:
	option.rendering_func = SHADED;
	break;
      case 2:
	option.rendering_func = BLENDED;
	break;
      default:
	break;
      }
      break;
    case 'f':
      printf("font file: %s\n",optarg);
      option.fontfile[0] = '\0';
      strncat(option.fontfile, optarg, sizeof(option.fontfile)-1);
      break;
    default:
      Usage();
      return 1;
    }
  }

  argc -= optind;
  argv += optind;
  if (argc == 0) {
    Usage();
    return 1;
  }
  option.filename[0] = '\0';
  strncat(option.filename, argv[0], sizeof(option.filename)-1);

  if (!Init()) {
    printf("error: init failed\n");
    return 1;
  }

  FillScreen(option.bg_color);
  TextSurface = TTF_RenderUTF8_Blended(font, "Loading...", option.fg_color);
  dest.x = 100;
  dest.y = 50;
  SDL_BlitSurface(TextSurface, NULL, gScreenSurface, &dest);
  SDL_FreeSurface(TextSurface);
  SDL_Flip(gScreenSurface);
  switch (option.text_mode) {
  case MODE_ALPHA:
    if (!LoadFile_alpha()) {
      printf("error: load failed\n");
      return 1;
    }
    break;
  case MODE_KANJI:
    if (!LoadFile_kanji()) {
      printf("error: load failed\n");
      return 1;
    }
    break;
  }
  MainLoop();
  End();

  return 0;
}
